/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file util/storage.c
 * @brief IO convenience methods
 * @author Christian Grothoff
 **/

#include "config.h"

#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


#if LINUX || CYGWIN
#include <sys/vfs.h>
#else
#ifdef SOMEBSD
#include <sys/param.h>
#include <sys/mount.h>
#else
#ifdef SOLARIS
#include <sys/types.h>
#include <sys/statvfs.h>
#else
#error FIXME: need to port statfs (how much space is left on the drive?)
#endif
#endif
#endif

#ifndef SOMEBSD
 #ifndef CYGWIN
 #include <wordexp.h>
 #endif
#endif

#include "util/storage.h"
#include "util/xmalloc.h"
#include "util/logging.h"
#include "util/returnvalues.h"

static void getSizeRec(char * filename,
		       char * dirname,
		       unsigned long long * size) {
  struct stat buf;  
  char * fn;

  if (filename == NULL)
    return;
  if (dirname != NULL) {
    fn = MALLOC(strlen(filename) + strlen(dirname) + 2);
    fn[0] = '\0';
    if (strlen(dirname) > 0) {
      strcat(fn, dirname);      
      if (dirname[strlen(dirname)-1] != '/')
	strcat(fn, "/"); /* add tailing / if needed */
    } else
      strcat(fn, "/"); 
    if (filename[0] == '/') /* if filename starts with a "/", don't copy it */
      strcat(fn, &filename[1]);
    else
      strcat(fn, filename);
  } else
    fn = STRDUP(filename);

  if (0 != stat(fn, &buf)) {
    LOG(LOG_DEBUG,
	"DEBUG: Can not stat %s (%s)\n",
	fn, 
	strerror(errno));
    FREE(fn);
    return;
  }
  *size += buf.st_size;
  if ( (S_ISDIR(buf.st_mode)) &&
       (!S_ISLNK(buf.st_mode)) ) {
    scanDirectory(fn,
		  (DirectoryEntryCallback)&getSizeRec,
		  size);
  }  
  FREE(fn);
}

/**
 * Get the number of blocks that are left on the partition that
 * contains the given file (for normal users).
 *
 * @param part a file on the partition to check
 * @return -1 on errors, otherwise the number of free blocks
 **/
long getBlocksLeftOnDrive(const char * part) {
#ifdef SOLARIS
  struct statvfs buf;

  if (0 == statvfs(part, &buf)) {
    return buf.f_bavail;
  } else {
    LOG(LOG_ERROR,
	"ERROR: statfs failed: %s\n",
	strerror(errno));
    return -1;
  }
#else
  struct statfs s;
  if (0 == statfs(part, &s)) {
    return s.f_bavail;
  } else {
    LOG(LOG_ERROR,
	"ERROR: statfs failed: %s\n",
	strerror(errno));
    return -1;
  }
#endif
}

/**
 * Get the size of the file (or directory)
 * of the given file (in bytes).
 **/
unsigned long long getFileSize(char * filename) {
  unsigned long long size;

  size = 0;
  LOG(LOG_EVERYTHING,
      "EVERYTHING: getting file size of %s\n",
      filename);
  getSizeRec(filename, "", &size);
  return size;
}


/**
 * Convert string to value ('755' for chmod-call)
 **/
static int atoo(char *s) {
  int n = 0;

  while ( ('0' <= *s) && (*s < '8') ) {
    n <<= 3;
    n += *s++ - '0';
  }
  return n;
}

/**
 * Assert that fil corresponds to a filename
 * (of a file that exists and that is not a directory).
 * @returns 1 if yes, 0 if not (will print an error
 * message in that case, too).
 **/
int assertIsFile(const char * fil) {
  struct stat filestat;
  int ret;

  ret = stat(fil, &filestat);
  if (ret != 0) {
    LOG(LOG_DEBUG,
	"DEBUG: Can not stat %s (%s).\n",
	fil,
	strerror(errno));
    return 0;
  }
  if (!S_ISREG(filestat.st_mode)) {
    LOG(LOG_WARNING,
	"WARNING: %s is not a regular file.\n",
	fil);
    return 0;
  }  
  if ( access(fil, R_OK) < 0 ) {
    LOG(LOG_WARNING,
	"WARNING: access failed (%s).\n",
	strerror(errno));
    return 0;
  }
  return 1;
}

/**
 * Complete filename (a la shell) from abbrevition.
 * @param fil the name of the file, may contain ~/ or 
 *        be relative to the current directory
 * @returns the full file name, 
 *          NULL is returned on error
 **/
char * expandFileName(const char * fil) {
  char * fm;
  char * fn;
  const char *fil_ptr;

  if (fil == NULL)
    return NULL;
  
  if (fil[0] == '/') {
    /* absolute path, just copy */
    return(STRDUP(fil));
  }
  else if (fil[0] == '~') {
    fm = getenv("HOME");
    if (fm == NULL) {
      /* keep it symbolic to show error to user! */
      fm = "$HOME";
    }

    /* do not copy '~' */
    fil_ptr = fil + 1;

	/* skip over dir seperator to be consistent */    
    if (fil_ptr[0] == '/') {
    	fil_ptr++;
    }
  } else {
    fil_ptr = fil;
    fm = getenv("PWD");
    if (fm == NULL)
      fm = "$PWD";
  }

  fn = MALLOC(strlen(fm) + 1 + strlen(fil_ptr) + 1);

  sprintf(fn, "%s/%s", fm, fil_ptr);

  return fn;
}

/**
 * Implementation of "mkdir -p"
 * @param dir the directory to create
 **/
void mkdirp(char * dir) {
  char * rdir;
  int len;
  int pos;

  rdir = expandFileName(dir); /* expand directory */
  len = strlen(rdir);
  pos = 1; /* skip heading '/' */
  while (pos <= len) {
    if ( (rdir[pos] == '/') || 
	 (pos == len) ) {
      rdir[pos] = '\0';
      if (0 != mkdir(rdir,
		     S_IRUSR | S_IWUSR | 
		     S_IXUSR | S_IRGRP | 
		     S_IXGRP | S_IROTH | 
		     S_IXOTH)) { /* 755 */
	if (errno != EEXIST)
	  LOG(LOG_ERROR,
	      "ERROR: could not mkdir %s: %s\n",
	      rdir,
	      strerror(errno));
      }
      rdir[pos] = '/';       
    }      
    pos++;
  }   
  FREE(rdir); 
}

/**
 * Read the contents of a binary file into a buffer.
 * @param fileName the name of the file, not freed,
 *        must already be expanded!
 * @param len the maximum number of bytes to read
 * @param result the buffer to write the result to
 * @return the number of bytes read on success, -1 on failure
 **/ 
int readFile(char * fileName,
	     int  len,
	     void * result) {
  /* open file, must exist, open read only */
  int handle;  
  int size;

  if ((fileName == NULL) || (result == NULL))
    return -1;
  handle = open(fileName,O_RDONLY,S_IRUSR);
  if (handle < 0)
    return -1; 
  size = read(handle, result, len);
  CLOSE(handle);
  return size;  
}

/**
 * Write a buffer to a file.
 * @param fileName the name of the file, NOT freed!
 * @param buffer the data to write
 * @param n number of bytes to write
 * @param mode permissions to set on the file
 **/ 
void writeFile(char * fileName, 
	       void * buffer,
	       int n,
	       char *mode) {
  int handle;
  /* open file, open with 600, create if not 
     present, otherwise overwrite */  
  if ((fileName == NULL) || (buffer == NULL))
    return;
  handle = open(fileName,
		O_CREAT|O_WRONLY,S_IRUSR|S_IWUSR);
  /* write the buffer take length from the beginning */
  if (n != write(handle, buffer, n)) 
    LOG(LOG_WARNING,
	"WARNING: Writing %d bytes to file %s failed!\n",
	n, 
	fileName);
  chmod(fileName, atoo(mode));
  CLOSE(handle);
}

/**
 * Build a filename from directory and filename, completing like the shell does
 * @param dir the name of the directory, may contain ~/ or other shell stuff. Will 
 *        NOT be freed!
 * @param fil the name of the file, will NOT be deallocated anymore!
 * @param result where to store the full file name (must be large enough!)
 **/
void buildFileName(char * dir,
		   HexName * fil,
		   char * result) {
  if ((dir == NULL) || (fil == NULL) || (result == NULL)) 
    errexit("buildFileName called with NULL in arguments");
  strcpy(result, dir);
  strcat(result, (char*)fil);
}

/**
 * Scan a directory for files. The name of the directory
 * must be expanded first (!).
 * @param dirName the name of the directory
 * @param callback the method to call for each file,
 *        can be NULL, in that case, we only count
 * @param data argument to pass to callback
 * @return the number of files found, -1 on error
 **/
int scanDirectory(char * dirName,
		  DirectoryEntryCallback callback,
		  void * data) {
  DIR * dinfo;
  struct dirent *finfo;
  struct stat istat;
  int count = 0;
 
  if (dirName == NULL) 
    return -1; 
  if (0 != stat(dirName, &istat)) {
    LOG(LOG_WARNING,
	"WARNING: Could not stat %s (%s)\n",
	dirName, 
	strerror(errno));
    return -1;
  }
  if (!S_ISDIR(istat.st_mode)) {
    LOG(LOG_ERROR,
	"ERROR: scanDirectory must be invoked on a directory (%s)!\n",
	dirName);
    return -1;
  }
  errno = 0;
  dinfo = opendir(dirName);
  if ((errno == EACCES) || (dinfo == NULL)) {
    LOG(LOG_WARNING,
	"WARNING: scanDirectory: %s (%s)\n",
	strerror(errno),
	(char*)dirName);
   return -1;
  }
  while ((finfo = readdir(dinfo)) != NULL) {
    if (finfo->d_name[0] == '.')
      continue;
    if (callback != NULL)
      callback(finfo->d_name,
	       dirName,
	       data);
    count++;
  }
  closedir(dinfo);
  return count;
}

/**
 * Callback for rm_minus_rf.
 **/
static void rmHelper(char * fil,
		     char * dir,
		     int * ok) {
  char * fn;
  fn = MALLOC(strlen(dir) + strlen(fil) + 2);
  sprintf(fn, "%s/%s", dir, fil);
  if (SYSERR == rm_minus_rf(fn))
    *ok = SYSERR;
  FREE(fn);
}

/**
 * Remove all files in a directory (rm -rf). Call with
 * caution.
 *
 *
 * @param fileName the file to remove
 * @return OK on success, SYSERR on error
 **/
int rm_minus_rf(char * fileName) {
  if (unlink(fileName) == 0)
    return OK;
  if (errno == EISDIR) {
    int ok;

    ok = OK;
    scanDirectory(fileName,
		  (DirectoryEntryCallback)&rmHelper,
		  &ok);
    if (ok == OK)
      if (0 != rmdir(fileName)) {
	LOG(LOG_WARNING,
	    "WARNING: could not remove %s: %s\n",
	    fileName, 
	    strerror(errno));
	ok = SYSERR;
      }
    return ok;
  } else {
    LOG(LOG_WARNING,
	"WARNING: could not remove %s: %s\n",
	fileName,
	strerror(errno));
    return SYSERR;
  }
}

void close_(int fd,
	    char * filename,
	    int linenumber) {
  if (0 != close(fd)) {
    LOG(LOG_INFO,
	"INFO: error closing file descriptor in %s:%d (%s)\n",
	filename,
	linenumber,
	strerror(errno));
  }
}

/* end of storage.c */
